@@ -66,7 +66,7 @@ def getServiceAndNode(uri):
         raise XMPPURIParseError("Empty URI path component")
 
     try:
-        jid = JID(entity)
+        service = JID(entity)
     except Exception, e:
         raise XMPPURIParseError("Invalid JID: %s" % e.message)
 
@@ -75,9 +75,17 @@ def getServiceAndNode(uri):
     try:
         nodeIdentifier = params['node'][0]
     except (KeyError, ValueError):
-        raise XMPPURIParseError("No node in query component of URI")
+        nodeIdentifier = ''
 
-    return jid, nodeIdentifier
+    return service, nodeIdentifier
+
+
+
+def getXMPPURI(service, nodeIdentifier):
+    """
+    Construct an XMPP URI from a service JID and node identifier.
+    """
+    return "xmpp:%s?;node=%s" % (service.full(), nodeIdentifier or '')
 
 
 
@@ -134,11 +142,12 @@ class CreateResource(resource.Resource):
         """
 
         def toResponse(nodeIdentifier):
-            uri = 'xmpp:%s?;node=%s' % (self.serviceJID.full(), nodeIdentifier)
+            uri = getXMPPURI(self.serviceJID, nodeIdentifier)
             stream = simplejson.dumps({'uri': uri})
-            return http.Response(responsecode.OK, stream=stream)
-
-        d = self.backend.create_node(None, self.owner)
+            contentType = http_headers.MimeType.fromString(MIME_JSON)
+            return http.Response(responsecode.OK, stream=stream,
+                                 headers={'Content-Type': contentType})
+        d = self.backend.createNode(None, self.owner)
         d.addCallback(toResponse)
         return d
 
@@ -184,7 +193,7 @@ class DeleteResource(resource.Resource):
                     "Malformed XMPP URI: %s" % failure.value.message)
 
         d = getNode()
-        d.addCallback(self.backend.delete_node, self.owner)
+        d.addCallback(self.backend.deleteNode, self.owner)
         d.addCallback(respond)
         d.addErrback(trapNotFound)
         d.addErrback(trapXMPPURIParseError)
@@ -237,9 +246,11 @@ class PublishResource(resource.Resource):
         """
 
         def toResponse(nodeIdentifier):
-            uri = 'xmpp:%s?;node=%s' % (self.serviceJID.full(), nodeIdentifier)
+            uri = getXMPPURI(self.serviceJID, nodeIdentifier)
             stream = simplejson.dumps({'uri': uri})
-            return http.Response(responsecode.OK, stream=stream)
+            contentType = http_headers.MimeType.fromString(MIME_JSON)
+            return http.Response(responsecode.OK, stream=stream,
+                                 headers={'Content-Type': contentType})
 
         def gotNode(nodeIdentifier, payload):
             item = Item(id='current', payload=payload)
@@ -252,7 +263,7 @@ class PublishResource(resource.Resource):
                 jid, nodeIdentifier = getServiceAndNode(request.args['uri'][0])
                 return defer.succeed(nodeIdentifier)
             else:
-                return self.backend.create_node(None, self.owner)
+                return self.backend.createNode(None, self.owner)
 
         def doPublish(payload):
             d = getNode()
@@ -286,10 +297,12 @@ class ListResource(resource.Resource):
 
     def render(self, request):
         def responseFromNodes(nodeIdentifiers):
-            return http.Response(responsecode.OK,
-                                 stream=simplejson.dumps(nodeIdentifiers))
+            stream = simplejson.dumps(nodeIdentifiers)
+            contentType = http_headers.MimeType.fromString(MIME_JSON)
+            return http.Response(responsecode.OK, stream=stream,
+                                 headers={'Content-Type': contentType})
 
-        d = self.service.get_nodes()
+        d = self.service.getNodes()
         d.addCallback(responseFromNodes)
         return d
 
@@ -325,8 +338,10 @@ def extractAtomEntries(items):
 
     return atomEntries
 
+
+
 def constructFeed(service, nodeIdentifier, entries, title):
-    nodeURI = 'xmpp:%s?;node=%s' % (service.full(), nodeIdentifier)
+    nodeURI = getXMPPURI(service, nodeIdentifier)
     now = strftime("%Y-%m-%dT%H:%M:%SZ", gmtime())
 
     # Collect the received entries in a feed
@@ -340,6 +355,8 @@ def constructFeed(service, nodeIdentifier, entries, title):
 
     return feed
 
+
+
 class RemoteSubscriptionService(service.Service, PubSubClient):
     """
     Service for subscribing to remote XMPP Publish-Subscribe nodes.
@@ -348,29 +365,34 @@ class RemoteSubscriptionService(service.Service, PubSubClient):
     to with the received items in notifications.
     """
 
-    def __init__(self, jid):
+    def __init__(self, jid, storage):
         self.jid = jid
-
-
-    def startService(self):
-        self.callbacks = {}
+        self.storage = storage
 
 
     def trapNotFound(self, failure):
         failure.trap(StanzaError)
-        if not failure.value.condition == 'item-not-found':
-            raise failure
-        raise error.NodeNotFound
+
+        if failure.value.condition == 'item-not-found':
+            raise error.NodeNotFound()
+        else:
+            return failure
 
 
     def subscribeCallback(self, jid, nodeIdentifier, callback):
+        """
+        Subscribe a callback URI.
 
-        def newCallbackList(result):
-            callbackList = set()
-            self.callbacks[jid, nodeIdentifier] = callbackList
-            return callbackList
+        This registers a callback URI to be called when a notification is
+        received for the given node.
 
-        def callbackForLastItem(items, callback):
+        If this is the first callback registered for this node, the gateway
+        will subscribe to the node. Otherwise, the most recently published item
+        for this node is retrieved and, if present, the newly registered
+        callback will be called with that item.
+        """
+
+        def callbackForLastItem(items):
             atomEntries = extractAtomEntries(items)
 
             if not atomEntries:
@@ -379,40 +401,49 @@ class RemoteSubscriptionService(service.Service, PubSubClient):
             self._postTo([callback], jid, nodeIdentifier, atomEntries[0],
                          'application/atom+xml;type=entry')
 
-        try:
-            callbackList = self.callbacks[jid, nodeIdentifier]
-        except KeyError:
-            d = self.subscribe(jid, nodeIdentifier, self.jid)
-            d.addCallback(newCallbackList)
-        else:
-            d = self.items(jid, nodeIdentifier, 1)
-            d.addCallback(callbackForLastItem, callback)
-            d.addCallback(lambda _: callbackList)
+        def subscribeOrItems(hasCallbacks):
+            if hasCallbacks:
+                d = self.items(jid, nodeIdentifier, 1)
+                d.addCallback(callbackForLastItem)
+            else:
+                d = self.subscribe(jid, nodeIdentifier, self.jid)
 
-        d.addCallback(lambda callbackList: callbackList.add(callback))
-        d.addErrback(self.trapNotFound)
+            d.addErrback(self.trapNotFound)
+            return d
+
+        d = self.storage.hasCallbacks(jid, nodeIdentifier)
+        d.addCallback(subscribeOrItems)
+        d.addCallback(lambda _: self.storage.addCallback(jid, nodeIdentifier,
+                                                         callback))
         return d
 
 
     def unsubscribeCallback(self, jid, nodeIdentifier, callback):
-        try:
-            callbackList = self.callbacks[jid, nodeIdentifier]
-            callbackList.remove(callback)
-        except KeyError:
-            return defer.fail(error.NotSubscribed())
+        """
+        Unsubscribe a callback.
 
-        if not callbackList:
-            self.unsubscribe(jid, nodeIdentifier, self.jid)
+        If this was the last registered callback for this node, the
+        gateway will unsubscribe from node.
+        """
 
-        return defer.succeed(None)
+        def cb(last):
+            if last:
+                return self.unsubscribe(jid, nodeIdentifier, self.jid)
+
+        d = self.storage.removeCallback(jid, nodeIdentifier, callback)
+        d.addCallback(cb)
+        return d
 
 
-    def itemsReceived(self, recipient, service, nodeIdentifier, items):
+    def itemsReceived(self, event):
         """
         Fire up HTTP client to do callback
         """
 
-        atomEntries = extractAtomEntries(items)
+        atomEntries = extractAtomEntries(event.items)
+        service = event.sender
+        nodeIdentifier = event.nodeIdentifier
+        headers = event.headers
 
         # Don't notify if there are no atom entries
         if not atomEntries:
@@ -428,19 +459,31 @@ class RemoteSubscriptionService(service.Service, PubSubClient):
 
         self.callCallbacks(service, nodeIdentifier, payload, contentType)
 
+        if 'Collection' in headers:
+            for collection in headers['Collection']:
+                nodeIdentifier = collection or ''
+                self.callCallbacks(service, nodeIdentifier, payload,
+                                   contentType)
 
-    def deleteReceived(self, recipient, service, nodeIdentifier):
+
+    def deleteReceived(self, event):
         """
         Fire up HTTP client to do callback
         """
 
+        service = event.sender
+        nodeIdentifier = event.nodeIdentifier
         self.callCallbacks(service, nodeIdentifier, eventType='DELETED')
 
 
     def _postTo(self, callbacks, service, nodeIdentifier,
                       payload=None, contentType=None, eventType=None):
+
+        if not callbacks:
+            return
+
         postdata = None
-        nodeURI = 'xmpp:%s?;node=%s' % (service.full(), nodeIdentifier)
+        nodeURI = getXMPPURI(service, nodeIdentifier)
         headers = {'Referer': nodeURI.encode('utf-8'),
                    'PubSub-Service': service.full().encode('utf-8')}
 
@@ -462,16 +505,20 @@ class RemoteSubscriptionService(service.Service, PubSubClient):
         for callbackURI in callbacks:
             reactor.callLater(0, postNotification, callbackURI)
 
+
     def callCallbacks(self, service, nodeIdentifier,
                             payload=None, contentType=None, eventType=None):
-        try:
-            callbacks = self.callbacks[service, nodeIdentifier]
-        except KeyError:
-            return
 
-        self._postTo(callbacks, service, nodeIdentifier, payload, contentType,
-                     eventType)
+        def eb(failure):
+            failure.trap(error.NoCallbacks)
 
+            # No callbacks were registered for this node. Unsubscribe?
+
+        d = self.storage.getCallbacks(service, nodeIdentifier)
+        d.addCallback(self._postTo, service, nodeIdentifier, payload,
+                                    contentType, eventType)
+        d.addErrback(eb)
+        d.addErrback(log.err)
 
 
 
@@ -496,6 +543,8 @@ class RemoteSubscribeBaseResource(resource.Resource):
                 (responsecode.FORBIDDEN, "Node not found"),
             error.NotSubscribed:
                 (responsecode.FORBIDDEN, "No such subscription found"),
+            error.SubscriptionExists:
+                (responsecode.FORBIDDEN, "Subscription already exists"),
     }
 
     def __init__(self, service):
@@ -572,6 +621,7 @@ class RemoteItemsResource(resource.Resource):
     def __init__(self, service):
         self.service = service
 
+
     def render(self, request):
         try:
             maxItems = int(request.args.get('max_items', [0])[0]) or None
@@ -580,14 +630,16 @@ class RemoteItemsResource(resource.Resource):
                     "The argument max_items has an invalid value.")
 
         try:
-            jid, nodeIdentifier = getServiceAndNode(request.args['uri'][0])
+            uri = request.args['uri'][0]
         except KeyError:
             return http.StatusResponse(responsecode.BAD_REQUEST,
                     "No URI for the remote node provided.")
+
+        try:
+            jid, nodeIdentifier = getServiceAndNode(uri)
         except XMPPURIParseError:
             return http.StatusResponse(responsecode.BAD_REQUEST,
-                    "Malformed XMPP URI: %s" % failure.value.message)
-
+                    "Malformed XMPP URI: %s" % uri)
 
         def respond(items):
             """Create a feed out the retrieved items."""
@@ -703,6 +755,13 @@ class GatewayClient(service.Service):
         pass
 
 
+    def ping(self):
+        f = getPageWithFactory(self._makeURI(''),
+                               method='HEAD',
+                               agent=self.agent)
+        return f.deferred
+
+
     def create(self):
         f = getPageWithFactory(self._makeURI('create'),
                     method='POST',
