@@ -18,14 +18,15 @@ from idavoll import gateway
 AGENT = "Idavoll Test Script"
 NS_ATOM = "http://www.w3.org/2005/Atom"
 
-entry = domish.Element((NS_ATOM, 'entry'))
-entry.addElement("id", content="urn:uuid:1225c695-cfb8-4ebb-aaaa-80da344efa6a")
-entry.addElement("title", content="Atom-Powered Robots Run Amok")
-entry.addElement("author").addElement("name", content="John Doe")
-entry.addElement("content", content="Some text.")
+TEST_ENTRY = domish.Element((NS_ATOM, 'entry'))
+TEST_ENTRY.addElement("id",
+                      content="urn:uuid:1225c695-cfb8-4ebb-aaaa-80da344efa6a")
+TEST_ENTRY.addElement("title", content="Atom-Powered Robots Run Amok")
+TEST_ENTRY.addElement("author").addElement("name", content="John Doe")
+TEST_ENTRY.addElement("content", content="Some text.")
 
 baseURI = "http://localhost:8086/"
-componentJID = "pubsub.localhost"
+componentJID = "pubsub"
 
 class GatewayTest(unittest.TestCase):
     timeout = 2
@@ -33,9 +34,26 @@ class GatewayTest(unittest.TestCase):
     def setUp(self):
         self.client = gateway.GatewayClient(baseURI)
         self.client.startService()
+        self.addCleanup(self.client.stopService)
+
+        def trapConnectionRefused(failure):
+            from twisted.internet.error import ConnectionRefusedError
+            failure.trap(ConnectionRefusedError)
+            raise unittest.SkipTest("Gateway to test against is not available")
+
+        def trapNotFound(failure):
+            from twisted.web.error import Error
+            failure.trap(Error)
+
+        d = self.client.ping()
+        d.addErrback(trapConnectionRefused)
+        d.addErrback(trapNotFound)
+        return d
+
 
     def tearDown(self):
-        self.client.stopService()
+        return self.client.stopService()
+
 
     def test_create(self):
 
@@ -51,7 +69,7 @@ class GatewayTest(unittest.TestCase):
         def cb(response):
             self.assertIn('uri', response)
 
-        d = self.client.publish(entry)
+        d = self.client.publish(TEST_ENTRY)
         d.addCallback(cb)
         return d
 
@@ -62,7 +80,7 @@ class GatewayTest(unittest.TestCase):
 
         def cb1(response):
             xmppURI = response['uri']
-            d = self.client.publish(entry, xmppURI)
+            d = self.client.publish(TEST_ENTRY, xmppURI)
             d.addCallback(cb2, xmppURI)
             return d
 
@@ -74,7 +92,7 @@ class GatewayTest(unittest.TestCase):
         def cb(err):
             self.assertEqual('404', err.status)
 
-        d = self.client.publish(entry, 'xmpp:%s?node=test' % componentJID)
+        d = self.client.publish(TEST_ENTRY, 'xmpp:%s?node=test' % componentJID)
         self.assertFailure(d, error.Error)
         d.addCallback(cb)
         return d
@@ -105,7 +123,7 @@ class GatewayTest(unittest.TestCase):
             return d
 
         def cb2(xmppURI):
-            d = self.client.publish(entry, xmppURI)
+            d = self.client.publish(TEST_ENTRY, xmppURI)
             return d
 
 
@@ -116,6 +134,50 @@ class GatewayTest(unittest.TestCase):
         d.addCallback(cb2)
         return defer.gatherResults([d, self.client.deferred])
 
+
+    def test_subscribeTwiceGetNotification(self):
+
+        def onNotification1(data, headers):
+            d = client1.stopService()
+            d.chainDeferred(client1.deferred)
+
+        def onNotification2(data, headers):
+            d = client2.stopService()
+            d.chainDeferred(client2.deferred)
+
+        def cb(response):
+            xmppURI = response['uri']
+            d = client1.subscribe(xmppURI)
+            d.addCallback(lambda _: xmppURI)
+            return d
+
+        def cb2(xmppURI):
+            d = client2.subscribe(xmppURI)
+            d.addCallback(lambda _: xmppURI)
+            return d
+
+        def cb3(xmppURI):
+            d = self.client.publish(TEST_ENTRY, xmppURI)
+            return d
+
+
+        client1 = gateway.GatewayClient(baseURI, callbackPort=8088)
+        client1.startService()
+        client1.callback = onNotification1
+        client1.deferred = defer.Deferred()
+        client2 = gateway.GatewayClient(baseURI, callbackPort=8089)
+        client2.startService()
+        client2.callback = onNotification2
+        client2.deferred = defer.Deferred()
+
+        d = self.client.create()
+        d.addCallback(cb)
+        d.addCallback(cb2)
+        d.addCallback(cb3)
+        dl = defer.gatherResults([d, client1.deferred, client2.deferred])
+        return dl
+
+
     def test_subscribeGetDelayedNotification(self):
 
         def onNotification(data, headers):
@@ -124,7 +186,7 @@ class GatewayTest(unittest.TestCase):
         def cb(response):
             xmppURI = response['uri']
             self.assertNot(self.client.deferred.called)
-            d = self.client.publish(entry, xmppURI)
+            d = self.client.publish(TEST_ENTRY, xmppURI)
             d.addCallback(lambda _: xmppURI)
             return d
 
@@ -157,7 +219,7 @@ class GatewayTest(unittest.TestCase):
             xmppURI = response['uri']
             self.assertNot(client1.deferred.called)
             self.assertNot(client2.deferred.called)
-            d = self.client.publish(entry, xmppURI)
+            d = self.client.publish(TEST_ENTRY, xmppURI)
             d.addCallback(lambda _: xmppURI)
             return d
 
@@ -198,6 +260,32 @@ class GatewayTest(unittest.TestCase):
         return d
 
 
+    def test_subscribeRootGetNotification(self):
+
+        def onNotification(data, headers):
+            self.client.deferred.callback(None)
+
+        def cb(response):
+            xmppURI = response['uri']
+            jid, nodeIdentifier = gateway.getServiceAndNode(xmppURI)
+            rootNode = gateway.getXMPPURI(jid, '')
+
+            d = self.client.subscribe(rootNode)
+            d.addCallback(lambda _: xmppURI)
+            return d
+
+        def cb2(xmppURI):
+            return self.client.publish(TEST_ENTRY, xmppURI)
+
+
+        self.client.callback = onNotification
+        self.client.deferred = defer.Deferred()
+        d = self.client.create()
+        d.addCallback(cb)
+        d.addCallback(cb2)
+        return defer.gatherResults([d, self.client.deferred])
+
+
     def test_unsubscribeNonExisting(self):
         def cb(err):
             self.assertEqual('403', err.status)
@@ -214,7 +302,7 @@ class GatewayTest(unittest.TestCase):
             d = self.client.items(xmppURI)
             return d
 
-        d = self.client.publish(entry)
+        d = self.client.publish(TEST_ENTRY)
         d.addCallback(cb)
         return d
 
@@ -225,6 +313,6 @@ class GatewayTest(unittest.TestCase):
             d = self.client.items(xmppURI, 2)
             return d
 
-        d = self.client.publish(entry)
+        d = self.client.publish(TEST_ENTRY)
         d.addCallback(cb)
         return d
